//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: client.hh,v 1.2 2002/12/15 10:10:07 loreti Exp $
// -------------------------------------------------------------------
// Header file for TCP/IP clients
// -------------------------------------------------------------------

#ifndef CLIENT_HH_
#define CLIENT_HH_

#include <cstring>
#include "sockhelp.hh"

extern "C" {
  #include <errno.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
  #include <netdb.h>
  #include <unistd.h>
}

// Class "Client": encapsulates an Internet TCP/IP client.

class Client {
public:
  typedef int socket_type;
  static const socket_type INVALID;

private:
  // Private data:
  // - The socket domain (AF_UNIX, AF_INET, ...); currently, only
  //   AF_INET is supported.
  // - The listening port;
  // - The server address;
  // - The network address structure related to the connecting socket;
  // - The file descriptor of the connecting socket.

  int                domain_;
  Port               port_;
  Address            address_;
  struct sockaddr_in servAddr_;
  socket_type        sockFD_;

  // Private methods (used by the constructors)

  void setSocket(int);

public:
  // Constructors:
  // - From the values of domain and type for the socket; the server
  //   host name; and the service name and protocol for the port.
  // - From the values of domain and type for the socket; the server
  //   host name; and the service name for the port;
  // - From the values of domain and type for the socket; the server
  //   host name; and the port number.
  // Destructor: virtual; closes the active socket.

  Client(int, int, const std::string &, const std::string &,
         const std::string &);
  Client(int, int, const std::string &, const std::string &);
  Client(int, int, const std::string &, Port::value_type);
  ~Client() {
    if (sockFD_ != INVALID) close(sockFD_);
  }

  // Public methods:
  // - To access read-only the Port instance related to the connecting
  //   socket;
  // - To access read-only the Address structure related to the
  //   connecting host;
  // - To test for a current active connection;
  // - To get the host name and the number of the active port of the
  //   connected server (the latter, in host order);
  // - To read or write a single character to or from the connecting
  //   socket;
  // - To read or write an array of characters to or from the
  //   connecting socket;
  // - To read or write a C-string to or from the connecting socket.

  const Port &     getPort()           const { return port_; }
  const Address &  getAddress()        const { return address_; }
  bool             isConnected()       const { return sockFD_ >= 0; }
  std::string      getServerName()     const;
  Port::value_type getHostServerPort() const;

  int sockGet(char * pc) {
    return MySockets::sockGet(sockFD_, pc);
  }
  int sockPut(const char * pc) {
    return MySockets::sockPut(sockFD_, pc);
  }
  int sockRead(char * pc, size_t s) {
    return MySockets::sockRead(sockFD_, pc, s);
  }
  int sockWrite(const char * pc, size_t s) {
    return MySockets::sockWrite(sockFD_, pc, s);
  }
  int sockGets(char * pc, size_t s) {
    return MySockets::sockGets(sockFD_, pc, s);
  }
  int sockPuts(const char * pc) {
    return MySockets::sockPuts(sockFD_, pc);
  }
};

// Inline methods for the class Client

inline Client::Client(
  int                 domain,
  int                 type,
  const std::string & hostAddr,
  const std::string & name,
  const std::string & protocol
) : domain_(domain), port_(name, protocol),
    address_(hostAddr), sockFD_(INVALID) {
  setSocket(type);
}

inline Client::Client(
  int                 domain,
  int                 type,
  const std::string & hostAddr,
  const std::string & name
) : domain_(domain), port_(name),
    address_(hostAddr), sockFD_(INVALID) {
  setSocket(type);
}

inline Client::Client(
  int                 domain,
  int                 type,
  const std::string & hostAddr,
  Port::value_type    number
) : domain_(domain), port_(number),
    address_(hostAddr), sockFD_(INVALID) {
  setSocket(type);
}

inline Port::value_type Client::getHostServerPort() const {
  // Gets the number of the port related to the outgoing connection,
  // in host order

  return sockFD_ < 0 ? 0 : ntohs(servAddr_.sin_port);
}

#endif // CLIENT_HH_
