//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: ipcserver.cxx,v 1.1 2002/11/28 17:50:59 loreti Exp $
// -------------------------------------------------------------------
// A toy TCP/IP server that uses the "fastServer" class, defined and
// implemented in server.hh and server.cxx.  The fastServer class
// abstracts a server designed to handle fast transactions with the
// client; it just waits for a connection, then deals with the
// communication between server and client in the input loop.
// -------------------------------------------------------------------

#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include "signals.hh"
#include "server.hh"

using std::cerr;
using std::cout;
using std::endl;

// Global variables:
// - The number of connections handled so far

unsigned int nCon = 0;

sigFunction * sigInt(
 const int
) {
  // Called when the server process has been killed (or hitting CTRL-C
  // or by a SIGINT signal); all the active sockets will be closed by
  // the fastServer destructor.

  cout << "Interrupted; connections handled: " << nCon << '\n';
  exit(EXIT_SUCCESS);
}

int main(
  int    argc,
  char * argv[]
) {
  // Command line arguments

  if (argc != 2) {
    cerr << "Usage: " << argv[0] << " <port>\n";
    return EXIT_SUCCESS;
  }

  // Signal handlers

  if (MySignal::mySignal(SIGPIPE, SIG_IGN) == SIG_ERR) {
    cerr << "Couldn't establish SIGPIPE handler\n";
    return EXIT_FAILURE;
  }

  if (
    MySignal::mySignal(SIGINT, reinterpret_cast<sigFunction *>(sigInt))
      == SIG_ERR) {
    cerr << "Couldn't establish SIGINT handler\n";
    return EXIT_FAILURE;
  }

  try {
    // Initializes a listening socket

    fastServer s(AF_INET, SOCK_STREAM, atoi(argv[1]));
    cout << "Server: listening " << s.getListeningPort() << endl;

    // Indefinite loop (kill with CTRL-C, or sending SIGINT).
    // - Waits for a connecting client;
    // - an informative message is printed to stdout;
    // - a sequence of characters is read, until the connection is
    //   closed;
    // - after every received character, the total number of
    //   characters read so far is sent back to the client.

    for (;;) {
      s.waitForCall();

      ++nCon;
      cout << "A client has been connected from host "
           << s.getConnectorName() << ", port "
           << s.getHostConnectorPort() << endl;

      char nRead = 0;
      while (s.isConnected()) {
        char c;
        if (s.sockGet(&c) > 0) {
          ++nRead;
          cout << "Server: got a \"" << c << "\" - acknowledged with "
               << static_cast<int>(nRead) << endl;
          s.sockPut(&nRead);
        }
      }

      cout << "Server: disconnecting" << endl;
    }

  } catch (const std::exception & e) {
    cerr << "Exception - " << e.what() << '\n';
  }
}
