//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: server.cxx,v 1.1 2002/11/28 17:50:59 loreti Exp $
// -------------------------------------------------------------------
// The implementation of the methods and procedures defined in
// server.hh
// -------------------------------------------------------------------

#include <stdexcept>
#include "server.hh"

using std::runtime_error;
using std::string;

// Static values and non-inline methods for the pure virtual class
// "Server"

unsigned int              Server::nInstances_ = 0;
const Server::socket_type Server::INVALID     = -1;
const int                 Server::LISTENQ     = 5;

void Server::setSocket(
  int type
) {
  // Overall initialization of the listening socket.  For the first
  // thing, we need a socket...

  listenFD_ = socket(domain_, type, 0);
  if (listenFD_ < 0) {
    string s( strerror(errno) );
    s = "(socket) " + s;
    throw runtime_error(s);
  }

  // Sets the "reuse address" flag for the socket

  static const int one = 1;

  if (setsockopt(listenFD_, SOL_SOCKET, SO_REUSEADDR,
                 &one, sizeof(one)) != 0) {
    string s( strerror(errno) );
    s = "(setsockopt) " + s;
    throw runtime_error(s);
  }

  // Connects the socket to the given port

  struct sockaddr_in servaddr;

  std::memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family      = domain_;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port        = port_.getNetValue();

  if (bind(listenFD_,
           reinterpret_cast<struct sockaddr *>(&servaddr),
           sizeof(servaddr)) != 0) {
    string s( strerror(errno) );
    s = "(bind) " + s;
    throw runtime_error(s);
  }

  // Makes the socket a listening socket

  if (listen(listenFD_, LISTENQ) != 0) {
    string s( strerror(errno) );
    s = "(listen) " + s;
    throw runtime_error(s);
  }
}

string Server::getConnectorName() const {
  // Gets the name of the connecting host, if available

  static const size_t BUFLEN = 128;
  char                buffer[BUFLEN];

  if (connectFD_ < 0   ||
      inet_ntop(domain_, &cliAddr_.sin_addr,
                buffer, BUFLEN) == 0) {
    return "";
  }
  return buffer;
}

Port::value_type Server::getHostConnectorPort() const {
  // Gets the number of the port related to the incoming connection,
  // in host order

  return connectFD_ < 0 ? 0 : ntohs(cliAddr_.sin_port);
}

// Methods for the concrete class "fastServer".  It implements a
// TCP/IP server targeted to fast transactions; it does not fork, but
// handles in the input loop all the client <==> server communication.

void fastServer::waitForCall()
{
  size_t length = sizeof(cliAddr_);

  // In some cases the (blocking) "accept" call that follows may be
  // interrupted, e.g. by a signal: if so, accept() returns -1 and
  // errno is set to EINTR.  In that case, waitForCall reiterates the
  // call to accept; if not, the execution is aborted.

  do {
#if defined( __osf__ ) || defined( __sgi__ )
    connectFD_ = accept(listenFD_,
                        reinterpret_cast<struct sockaddr *>(&cliAddr_),
                        reinterpret_cast<int *>(&length));
#else
    connectFD_ = accept(listenFD_,
                        reinterpret_cast<struct sockaddr *>(&cliAddr_),
                        &length);
#endif // __osf__ or __sgi__

    if (connectFD_ < 0   &&   errno != EINTR) {
      string s( strerror(errno) );
      s = "(accept) " + s;
      close(connectFD_);
      throw runtime_error(s);
    }
  } while (connectFD_ < 0);
}

// Methods for the concrete class "slowServer".  It implements a
// TCP/IP server targeted to slow transactions; it fork internally, so
// that every client <==> server communication is handled by a
// separate process.

void slowServer::waitForCall()
{
  size_t length = sizeof(cliAddr_);

  while (connectFD_ < 0) {

    // This is the main loop: we wait here for a connection from a
    // client.  The file descriptor of the connecting socket related
    // to the client is returned by "accept"; see also the comments
    // about accept in the fastServer::waitForCall implementation.

    do {
#if defined( __osf__ ) || defined( __sgi__ )
      connectFD_ = accept(listenFD_,
                          reinterpret_cast<struct sockaddr *>(&cliAddr_),
                          reinterpret_cast<int *>(&length));
#else
      connectFD_ = accept(listenFD_,
                          reinterpret_cast<struct sockaddr *>(&cliAddr_),
                          &length);
#endif // __osf__ or __sgi__

      if (connectFD_ < 0   &&   errno != EINTR) {
        string s( strerror(errno) );
        s = "(accept) " + s;
        close(connectFD_);
        throw runtime_error(s);
      }
    } while (connectFD_ < 0);

    int newProc;

    if ((newProc = fork()) < 0) {
      // fork() error

      std::cerr << "(fork) " << strerror(errno) << '\n';
      close(connectFD_);
      connectFD_ = INVALID;

    } else if (newProc == 0) {
      // This is the child process: we close our copy of the listening
      // socket, then exit the main loop to handle the communication.

      close(listenFD_);
      listenFD_ = INVALID;

    } else {
      // This is the parent process: we close our copy of the
      // connecting socket, and continue looping waiting for another
      // client.

      close(connectFD_);
      connectFD_ = INVALID;

    } // if (fork)
  } // while (connectFD_)
}
