//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: server.hh,v 1.2 2002/12/15 10:10:07 loreti Exp $
// -------------------------------------------------------------------
// The definition of the pure virtual class "Server" that encapsulates
// all what is needed in a TCP/IP server except the part that handles
// the individual connection; and of two concrete classes derived from
// Server, i.e. "slowServer" and "fastServer".  The former class is
// designed for slow transactions: the server forks, and every client
// is handled by an individual process; the latter class is designed
// for fast transactions: the server just returns and handles the
// communication with the client in the input loop.
// -------------------------------------------------------------------

#ifndef SERVER_HH_
#define SERVER_HH_

#include <cstring>
#include "sockhelp.hh"

extern "C" {
  #include <errno.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
  #include <netdb.h>
  #include <unistd.h>
}

// Class "Server": encapsulates an Internet TCP/IP server.  Server is
// a PURE VIRTUAL class.

class Server {
public:
  // LISTENQ is the maximum number of pending requests admissible on
  // the socket

  typedef int              socket_type;
  static const socket_type INVALID;
  static const int         LISTENQ;

private:
  // Private data:
  // - The number of instances for the class (to roughly implement a
  //   "singleton" pattern);
  // - The socket domain (AF_UNIX, AF_INET, ...); currently, only
  //   AF_INET is supported.
  // - The listening port.

  static unsigned int nInstances_;
  int                 domain_;
  Port                port_;

protected:
  // Protected data:
  // - The file descriptor of the listening socket;
  // - The file descriptor of the connecting socket;
  // - The network address structure related to the connecting socket.

  socket_type        listenFD_;
  socket_type        connectFD_;
  struct sockaddr_in cliAddr_;

private:
  // Private methods (used by the constructors)

  void test();
  void setSocket(int);

public:
  // Constructors:
  // - From the values of domain and type for the socket; and the
  //   service name and protocol for the port.
  // - From the values of domain and type for the socket; and the
  //   service name for the port.
  // - From the values of domain and type for the socket; and the port
  //   number.
  // Destructor: virtual; closes the active sockets.

  Server(int, int, const std::string &, const std::string &);
  Server(int, int, const std::string &);
  Server(int, int, Port::value_type);
  virtual ~Server() {
    if (listenFD_ >= 0)  close(listenFD_);
    if (connectFD_ >= 0) close(connectFD_);
  }

  // Public methods:
  // - To wait for an incoming connection (PURE VIRTUAL);
  // - To test for a current active connection;
  // - To get the host name and the number of the active port of the
  //   connecting client (the latter, in host order);
  // - To access read-only the Port instance related to the listening
  //   socket;
  // - To read or write a single character to or from the connecting
  //   socket;
  // - To read or write an array of characters to or from the
  //   connecting socket;
  // - To read or write a C-string to or from the connecting socket.

  virtual void     waitForCall()                = 0;
  bool             isConnected()          const { return connectFD_ >= 0; }
  std::string      getConnectorName()     const;
  Port::value_type getHostConnectorPort() const;
  const Port &     getListeningPort()     const { return port_; }

  int sockGet(char * pc) {
    return MySockets::sockGet(connectFD_, pc);
  }
  int sockPut(const char * pc) {
    return MySockets::sockPut(connectFD_, pc);
  }
  int sockRead(char * pc, size_t s) {
    return MySockets::sockRead(connectFD_, pc, s);
  }
  int sockWrite(const char * pc, size_t s) {
    return MySockets::sockWrite(connectFD_, pc, s);
  }
  int sockGets(char * pc, size_t s) {
    return MySockets::sockGets(connectFD_, pc, s);
  }
  int sockPuts(const char * pc) {
    return MySockets::sockPuts(connectFD_, pc);
  }
};                              // Class Server

// Class fastServer: concrete class derived from Server

class fastServer : public Server {
public:
  fastServer(int d, int t, const std::string & n, const std::string & p)
    : Server(d, t, n, p) {}
  fastServer(int d, int t, const std::string & n)
    : Server(d, t, n) {}
  fastServer(int d, int t, Port::value_type n)
    : Server(d, t, n) {}
  ~fastServer() {}
  void waitForCall();
};                              // Class fastServer

// Class slowServer: concrete class derived from Server

class slowServer : public Server {
public:
  slowServer(int d, int t, const std::string & n, const std::string & p)
    : Server(d, t, n, p) {}
  slowServer(int d, int t, const std::string & n)
    : Server(d, t, n) {}
  slowServer(int d, int t, Port::value_type n)
    : Server(d, t, n) {}
  ~slowServer() {}
  void waitForCall();
};                              // Class slowServer

// Inline methods for the class Server

inline Server::Server(
  int                 domain,
  int                 type,
  const std::string & name,
  const std::string & protocol
) : domain_(domain), port_(name, protocol),
    listenFD_(INVALID), connectFD_(INVALID) {
  test();
  setSocket(type);
}

inline Server::Server(
  int                 domain,
  int                 type,
  const std::string & name
) : domain_(domain), port_(name),
    listenFD_(INVALID), connectFD_(INVALID) {
  test();
  setSocket(type);
}

inline Server::Server(
  int              domain,
  int              type,
  Port::value_type number
) : domain_(domain), port_(number),
    listenFD_(INVALID), connectFD_(INVALID) {
  test();
  setSocket(type);
}

inline void Server::test() {
  // - Only a single instance of the class is allowed;
  // - Currently, the only supported domain is AF_INET;
  // - The supplied port must be valid.

  if (nInstances_++ != 0) {
    throw std::runtime_error("More than one server instance");
  }
  if (domain_ != AF_INET) {
    throw std::runtime_error("Domain not supported");
  }
  if (!port_.isValid()) {
    throw std::runtime_error("Invalid port");
  }
}

#endif // SERVER_HH_
