//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: sockhelp.cxx,v 1.2 2004/04/23 06:44:54 loreti Exp $
// -------------------------------------------------------------------
// Implementation of the methods and procedures defined in sockhelp.hh
// -------------------------------------------------------------------

#include <stdexcept>

extern "C" {
  #include <errno.h>
}

#include "sockhelp.hh"

using std::cerr;
using std::ostream;
using std::strerror;

// Static values and methods for the class "Port"

const Port::value_type Port::INVALID = 0;

void Port::getPortByName(
  const char * name,
  const char * proto
) {
  struct servent * pServ = getservbyname(name, proto);

  if (pServ) {
    number_ = pServ->s_port;
    name_   = pServ->s_name;
  } else {
    number_ = INVALID;
  }
}

void Port::out(
  ostream & o
) const {
  if (number_ == INVALID) {
    o << "(*INVALID*)";
  } else {
    if (!name_.empty()) {
      o << "(Service " << name_ << ", port ";
    } else {
      o << "(Port ";
    }
    o << ntohs(number_) << ')';
  }
}

// Methods for the class "Address"

Address::Address(
  const std::string & input
) {
#if !defined( __sun__ )
  if (inet_aton(input.c_str(), &addr_) != 0) {
    asciiAddr = input;
    return;
  }
#else
  // Solaris does not have inet_aton, but the (obsolete) inet_addr

  addr_.s_addr = inet_addr(input.c_str());
  if (addr_.s_addr != static_cast<unsigned long>(-1)) {
    asciiAddr = input;
    return;
  }
#endif // __sun__

  struct hostent * pHE = gethostbyname(input.c_str());
  if (pHE) {
    std::memcpy(&addr_, *pHE->h_addr_list, sizeof(addr_));
    asciiAddr = inet_ntoa(addr_);
  } else {
    asciiAddr = "";
  }
}

void Address::out(
  ostream & o
) const {
  if (asciiAddr.empty()) o << "(*INVALID*)";
  else {
    std::ios_base::fmtflags flags = o.flags();
    o << "(0x"
      << std::hex << std::uppercase << ntohl(addr_.s_addr)
      << " = " << asciiAddr << ')';
    o.setf(flags);
  }
}

// Static constants ans methods for the MySockets namespace (generic
// I/O through sockets)

int MySockets::sockRead(
  socket_type & socket,
  char *        buf,
  size_t        count
) {
  // This is just like the read() UNIX system call, except that it
  // will make sure that all your data goes through the socket.

  size_t bytes_read = 0;
  int    this_read;

  while (bytes_read < count) {
    // Reiterates the call when interrupted by some signal while
    // reading (EINTR I/O error))

    do {
      this_read = read(socket, buf, count - bytes_read);
    } while (this_read < 0   &&   errno == EINTR);

    // Here the value of this_read may be:
    // +N for a successful reading of N bytes;
    //  0 if an and-of-file has been received;
    // -1 if a read error has occurred.

    if (this_read <= 0) {
      int save_errno = errno;
      close(socket);
      socket = INVALID;

      if (this_read < 0) {      // Read error
        errno = save_errno;
        cerr << "(read) " << strerror(errno) << '\n';
        return this_read;
      } else  {                 // EOF
        return bytes_read;
      }
    }

    bytes_read += this_read;
    buf        += this_read;
  }

  return count;
}

int MySockets::sockWrite(
  socket_type & socket,
  const char *  buf,
  size_t        count
) {
  // This is just like the write() UNIX system call, except that it
  // will make sure that all your data goes through the socket.

  size_t bytes_sent = 0;
  int    this_write;

  while (bytes_sent < count) {
    do {
      this_write = write(socket, buf, count - bytes_sent);
    } while (this_write < 0   &&   errno == EINTR);

    if (this_write <= 0) {
      if (this_write < 0) {
        cerr << "(write) " << strerror(errno) << '\n';
      }
      close(socket);
      socket = INVALID;
      return this_write;
    }

    bytes_sent += this_write;
    buf        += this_write;
  }

  return count;
}

int MySockets::sockGet(
  socket_type & socket,
  char *        pc
) {
  // As sock_read, but reads a single byte

  int n;

  do {
    n = read(socket, pc, 1);
  } while (n < 0   &&   errno == EINTR);

  if (n != 1) {
    if (n < 0) {
      cerr << "(read) " << strerror(errno) << '\n';
    }
    close(socket);
    socket = INVALID;
  }
  return n;
}

int MySockets::sockPut(
  socket_type & socket,
  const char *  pc
) {
  // As sock_write, but writes a single byte

  int n;

  do {
    n = write(socket, pc, 1);
  } while (n < 0   &&   errno == EINTR);

  if (n != 1) {
    if (n < 0) {
      cerr << "(write) " << strerror(errno) << '\n';
    }
    close(socket);
    socket = INVALID;
  }
  return n;
}

int MySockets::sockGets(
  socket_type & socket,
  char *        str,
  size_t        count
) {
  // This function reads from a socket, until it receives a line-feed
  // ('\n') character and filling the character array "str" up to the
  // maximum size of "count" bytes.
  //
  // NOTE that if the readout line exceeds the length of "count", the
  // extra data will be READ AND DISCARDED.  *You have been warned*.

  int    bytes_read;
  size_t total_count      = 0;
  char * current_position = str;
  char   last_read        = '\0';

  while (last_read != '\n') {
    if ((bytes_read = read(socket, &last_read, 1)) <= 0) {
      if (bytes_read < 0) {
        cerr << "(read) " << strerror(errno) << '\n';
      }
      close(socket);
      socket = INVALID;
      return bytes_read;
    }

    if (total_count < count   &&   last_read != '\n') {
      *current_position++ = last_read;
      total_count++;
    }
  }

  if (count > 0) {
    *current_position = '\0';
  }
  return total_count;
}

int MySockets::sockPuts(
  socket_type & socket,
  const char *  str
) {
  // This function writes a character array to a socket.

  return sockWrite(socket, str, std::strlen(str));
}

// Digital Unix does not have a working implementation of inet_ntop

#if defined( __osf__ )

const char *inet_ntop(
  int         af,
  const void *src,
  char       *dst,
  size_t      cnt
) {
  if (af != AF_INET) {
    errno = EAFNOSUPPORT;
    return 0;
  } else {
    static const size_t   BUFLEN = 128;
    char                  temp[BUFLEN];
    const unsigned char * puc = static_cast<const unsigned char *>(src);
    size_t                n;

    n = std::sprintf(temp, "%u.%u.%u.%u", puc[0], puc[1], puc[2], puc[3]);
    if (n >= cnt) {
      errno = ENOSPC;
      return 0;
    } else {
      std::strcpy(dst, temp);
      return dst;
    }
  }
}

#endif // __osf__
