//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: sockhelp.hh,v 1.2 2002/12/05 10:42:34 loreti Exp $
// -------------------------------------------------------------------
// Procedures that deal with ports, TCP/IP addresses and socket I/O;
// designed to encapsulate the different implementations found on
// several operating systems - and to present to the user a consistent
// interface.
// -------------------------------------------------------------------

#ifndef SOCKHELP_HH_
#define SOCKHELP_HH_

#include <iostream>
#include <string>

extern "C" {
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
  #include <netdb.h>
  #include <unistd.h>
}

// Class "Port"; encapsulates a port number.

class Port {
public:
  typedef unsigned short int value_type;
  static const value_type    INVALID;

private:
  // Data:
  // - The port number, stored IN NETWORK ORDER;
  // - The name of the related service (if known).

  value_type  number_;
  std::string name_;

  // Private method: to get the port number from service and protocol

  void getPortByName(const char *, const char *);

public:
  // Constructors:
  // - Default, initializes the instance to an invalid value;
  // - From two strings containing the service name and the protocol;
  // - From a string containing the service name (/etc/services will
  //   be queried for the default protocol);
  // - From a straight port number;
  // Destructor: empty.

  Port() : number_(INVALID), name_("") {}
  Port(const std::string & name, const std::string & protocol)
    : name_("") { getPortByName(name.c_str(), protocol.c_str()); }
  Port(const std::string & name)
    : name_("") { getPortByName(name.c_str(), 0); }
  Port(const value_type number)
    : name_("") { number_ = htons(number); }
  ~Port() {}

  // Accessors:
  // - To test for a valid port number;
  // - To get the port number, in host OR in network order;
  // - To get the service name (if known);
  // - A public method for the printout (the port number will be
  //  printed in HOST order).

  bool                isValid()      const { return number_ != INVALID; }
  value_type          getNetValue()  const { return number_; }
  value_type          getHostValue() const { return ntohs(number_); }
  const std::string & getName()      const { return name_; }

  void out(std::ostream & o = std::cout) const;
};                              // Class Port

// A procedure that prints the internals of a Port instance into an
// output stream

inline std::ostream & operator << (
  std::ostream & o,
  const Port   & p
) {
  p.out(o);
  return o;
}

// Class "Address"; encapsulates an Internet address.

class Address {
private:
  // Data:
  // - A struct in_addr, that internally contains the TCP/IP address
  //   stored IN NETWORK ORDER;
  // - The translation of the above address into a string (written in
  //   number-and-dot notation).

  struct in_addr addr_;
  std::string    asciiAddr;

public:
  // Constructors:
  // - Default, initializes the instance to an invalid value;
  // - From a string containing the host name; or in number-and-dot
  //   notation, or in the more human-readable host-domain format.
  // Destructor: empty.

  Address() : asciiAddr("") {}
  Address(const std::string &);
  ~Address() {}

  // Accessors:
  // - To test for a valid address;
  // - To get the Internet address, as an unsigned long int IN NETWORK
  //   ORDER or as a string written in number-and-dot notation;
  // - A public method for the printout (the host number will be
  //   printed in HOST order).

  bool                isValid()      const { return !asciiAddr.empty(); }
  unsigned long       getNetAddr()   const { return addr_.s_addr; }
  unsigned long       getHostAddr()  const { return ntohl(addr_.s_addr); }
  const std::string & getAsciiAddr() const { return asciiAddr; }

  void out(std::ostream & o = std::cout) const;
};                              // Class Address

// A procedure that prints the internals of an Address instance into
// an output stream

inline std::ostream & operator << (
  std::ostream  & o,
  const Address & a
) {
  a.out(o);
  return o;
}

namespace MySockets {
  typedef int socket_type;
  static const socket_type INVALID = -1;

  int sockGet(socket_type &, char *);
  int sockPut(socket_type &, const char *);
  int sockRead(socket_type &, char *, size_t);
  int sockWrite(socket_type &, const char *, size_t);
  int sockGets(socket_type &, char *, size_t);
  int sockPuts(socket_type &, const char *);
} // namespace MySockets

// Digital Unix does not have a working implementation of inet_ntop; a
// quick-and-dirty replacement is in sockhelp.cxx

#if defined( __osf__ )
  #include <cstring>
  #include <cstdio>

  const char *inet_ntop(int, const void *, char *, size_t);
#endif // __osf__

#endif // SOCKHELP_HH_
