//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// $Id: toyserver.cxx,v 1.4 2006/09/11 12:27:04 loreti Exp $
// -------------------------------------------------------------------
// A toy TCP/IP server that uses the "slowServer" class, defined and
// implemented in server.hh and server.cxx.  The slowServer class
// abstracts a server designed to handle slow transactions with the
// clients; it waits for a connection, then forks - so that every
// communication between the server and a client will be handled by a
// different process.
//
// If the preprocessor macro DAEMON is defined, toyserver sets himself
// up as a running daemon.
// -------------------------------------------------------------------

#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include "signals.hh"
#include "server.hh"

using std::cerr;
using std::string;

#if defined(DAEMON)
  #include "daemonize.hh"

  string progName;

  string baseName(const char *);
#endif

extern "C" {
  #include <sys/wait.h>
}

sigFunction * sigInt(
 const int
) {
  // Called when the server process has been killed (or hitting CTRL-C
  // or by a SIGINT signal); all the active sockets will be closed by
  // the slowServer destructor.

  std::cout << "Interrupted\n";
  exit(EXIT_SUCCESS);
}

sigFunction * sigChild(
 const int
) {
  // Called when the forked children terminate

  int status;

  while (waitpid(-1, &status, WNOHANG) > 0) { }
  return 0;                     // To shut up 'lint'
}

int main(
  int    argc,
  char * argv[]
) {
#if defined(DAEMON)
  progName = baseName(argv[0]);
#endif

  // Command line arguments

  if (argc != 2) {
    cerr << "Usage: " << argv[0] << " <port>\n";
    return EXIT_SUCCESS;
  }

  // Signal handlers

  if (MySignal::mySignal(SIGPIPE, SIG_IGN) == SIG_ERR) {
    cerr << "Couldn't establish SIGPIPE handler\n";
    return EXIT_FAILURE;
  }

  if (
    MySignal::mySignal(SIGINT, reinterpret_cast<sigFunction *>(sigInt))
      == SIG_ERR) {
    cerr << "Couldn't establish SIGINT handler\n";
    return EXIT_FAILURE;
  }

  if (
    MySignal::mySignal(SIGCHLD, reinterpret_cast<sigFunction *>(sigChild))
      == SIG_ERR) {
    cerr << "Couldn't establish SIGCHLD handler\n";
    return EXIT_FAILURE;
  }

  try {
#if defined(DAEMON)
    MyDaemon::daemonize(progName);
#endif

    // Initializes a listening socket

    slowServer s(AF_INET, SOCK_STREAM, atoi(argv[1]));

    // Indefinite loop (kill with CTRL-C or sending SIGINT).
    // - Waits for a connecting client;
    // - a welcome message is sent to the client;
    // - a sequence of characters terminated by '\n' is read from the
    //   client;
    // - the same characters, translated to uppercase, are sent back
    //   to the client;
    // - The last two steps are repeated, until the connection is
    //   closed.

    s.waitForCall();
    s.sockPuts("Welcome to the upper case server.\n");

    while (s.isConnected()) {
      static const size_t BUFLEN = 512;
      char buffer[BUFLEN];

      if (s.sockGets(buffer, BUFLEN) > 0) {
        char * pc;

        for (pc = buffer;   *pc != '\0';   ++pc) {
          *pc = std::toupper(static_cast<unsigned char>(*pc));
        }
        *pc++ = '\n';
        *pc = '\0';

        s.sockPuts(buffer);
      }
    }
  } catch (const std::exception & e) {
    cerr << "Exception - " << e.what() << '\n';
  }
}

#if defined(DAEMON)

string baseName(
  const char * pc
) {
  // Strips the (eventual) path information from the program name (in
  // pc, i.e. argv[0]), returning the result as a std::string; if no
  // file name is given, returns an empty string.

  char * p = std::strrchr(pc, '/');
  if (p == 0) {
    return pc;
  } else {
    return ++p;
  }
}

#endif // DAEMON
