#include <ctype.h>              /* #include's: standard library */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <unistd.h>             /* #include's: standard POSIX */
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#if defined(__linux__)          /* #include's: OS dependent */
  #include <bits/sigaction.h>
  #include <bits/sigset.h>
#endif

#if defined(__linux__) || defined(__sgi__)
int sigemptyset();              /* Black magic?  Prototypes not seen,  */
int sigaction();                /*   also with proper header included. */
#endif

#if defined(__sun__) || defined(__sgi__)
  #include <sys/signal.h>
#endif

#if defined(__sun__) || defined(__osf__)
  const char *inet_ntop(int, const void *, char *, size_t);
#endif

/**
 | Defined constants.
 |
 | - BUFLEN:  length of the all-purpose buffers
 | - PORT:    default value of the well-known port used
 | - LISTENQ: maximum number of pending requests
**/

#define BUFLEN  128
#define PORT    1234
#define LISTENQ 5

/**
 | Shortcuts.
 |
 | sigFun: defined as a function taking a const int and returning void
**/

typedef void sigFun(const int);

/**
 | Local procedures: prototypes
**/

static void    main_loop();
static sigFun *my_signal(const int, sigFun *);
static void    sig_int(const int);
static void    usage(const char *);

/**
 | Global variables.
 |
 | listenfd: file descriptor of the listening socket (will be
 |           negative if errors occurred)
**/

int listenfd = -1;

int main(
  int   argc,
  char *argv[]
) {
  int retVal = EXIT_FAILURE;    /* Value returned to the OS */
  short port = PORT;            /* Listening port */

  /**
   | Tries to decode the (optional) port address contained in argv[1];
   | the argument is parsed by atoi (quick and dirty).
  **/

  switch (argc) {
    case 1:
      break;

    case 2:
      if (isdigit((unsigned char) *(argv[1]))) {
        port = atoi(argv[1]);
        break;
      } /* Intentional fallthrough */

    default:
      usage(argv[0]);
      return EXIT_FAILURE;
  }

  /**
   | Establishes the signal handlers: SIGPIPE is ignored; after an
   | interrupt the final cleanup is performed.
  **/
  
  if (my_signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
    perror("sigaction(SIGPIPE)");
  } else {
    if (my_signal(SIGINT, sig_int) == SIG_ERR) {
      perror("sigaction(SIGINT)");
    } else {

      /**
       | Gets a socket; sets the 'reuse address' flag; connects the
       | socket to the well-known port; makes the socket a listening
       | socket.
      **/
       
      if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        perror("socket");
      } else {
        const int one = 1;

        if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR,
                       &one, sizeof(one)) != 0) {
          perror("setsockopt");
        } else {
          struct sockaddr_in servaddr;

          memset(&servaddr, 0, sizeof(servaddr));
          servaddr.sin_family      = AF_INET;
          servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
          servaddr.sin_port        = htons(port);

          if (bind(listenfd, &servaddr, sizeof(servaddr)) != 0) {
            perror("bind");
          } else {
            if (listen(listenfd, LISTENQ) != 0) {
              perror("listen");
            } else {
              printf("Server: listening on port %hd\n", port);
              main_loop();
              retVal = EXIT_SUCCESS;
            } /* listen()        error ? */
          }   /* bind()          error ? */
        }     /* setsockopt()    error ? */
      }       /* socket()        error ? */
    }         /* signal(SIGINT)  error ? */
  }           /* signal(SIGPIPE) error ? */

  /**
   | If a file descriptor has been obtained, will be closed here
  **/

  if (listenfd >= 0) {
    if (close(listenfd) != 0) {
      perror("close");
      return EXIT_FAILURE;
    }
  }

  return retVal;
}

static void main_loop(void)
{
  struct sockaddr_in cliAddr;
  size_t             length;
  int                connfd;

  /**
   | The server loops forever; you have to interrupt it with CTRL-C.
  **/

  for (;;) {
    length = sizeof(cliAddr);
    if ((connfd = accept(listenfd, &cliAddr, &length)) < 0) {
      perror("accept");
    } else {
      char buf[BUFLEN];

      /**
       | When a client connects:
       | - an informative message is printed on stdout;
       | - a sequence of single characters is read, until the
       |   connection is closed;
       | - after every received character, the total number of
       |   characters read so far is sent back to the client.
      **/

      if (inet_ntop(AF_INET, &cliAddr.sin_addr, buf, BUFLEN) == 0) {
        perror("inet_ntop");
      } else {
        unsigned char nRead = 0;
        
        printf("Connected from %s, port %d\n", buf, ntohs(cliAddr.sin_port));
        
        while (connfd >= 0) {
          char c;
          int  n;

          do {
            n = read(connfd, &c, 1);
          } while (n < 0   &&   errno == EINTR);

          /**
           | Here n is:
           | -  1 for a successful read;
           | -  0 if an end-of-file has been received;
           | - -1 if a read error occurred.
          **/
           
          if (n != 1) {
            if (n < 0) perror("read");
            if (close(connfd) != 0) perror("connfd");
            connfd = -1;
          } else {
            nRead++;
              
            do {
              n = write(connfd, &nRead, 1);
            } while (n < 0   &&   errno == EINTR);

            /**
             | Here n is:
             | -  1 for a successful write;
             | - -1 if a write error occurred.
            **/
 
            if (n != 1) {
              perror("write");
              if (close(connfd) != 0) perror("connfd");
              connfd = -1;
            } else {
              printf("Server - got \"%c\", acknowledged with %d\n",
                     c, (int) nRead);
            } /* if (write OK) */
          } /* if (read OK) */
        } /* loop until connected */
      } /* if (inet_ntop) */
      puts("This connection has been closed");
    } /* accept */
  } /* for (;;) */
}

static void sig_int(
 const int sig
) {

/**
 | The server has been killed by CTRL-C
**/

  puts("Interrupted");
  close(listenfd);
  exit(EXIT_SUCCESS);
}

#if defined(__sun__)

/*****  See the comments in "sockhelp.c" for the *****/
/*****  details about the procedure my_signal(). *****/

static sigFun *my_signal(
  const int  signalNumber,
  sigFun    *fun
) {
  return signal(signalNumber, fun);
}

#else

static sigFun *my_signal(
  const int  signalNumber,
  sigFun    *fun
) {
  struct sigaction action, oldAction;

/**
 | - Declare "fun" as handler for "sigNr": no additional signals to
 |   be blocked when our signal handler is called;
 | - patch for SA_INTERRUPT (bad behavior of older SunOS 4.x);
 | - patch for SA_RESTART (SVR4, 4.4BSD).
**/

  sigemptyset(&action.sa_mask);
  action.sa_handler = fun;
  action.sa_flags   = 0;

  if (signalNumber == SIGALRM) {
#if defined(SA_INTERRUPT)
    action.sa_flags |= SA_INTERRUPT;
#endif
  } else {
#if defined(SA_RESTART)
    action.sa_flags |= SA_RESTART;
#endif
  }

  if (sigaction(signalNumber, &action, &oldAction) < 0) {
    return SIG_ERR;
  } else {
    return oldAction.sa_handler;
  }
}

#endif /* __sun__ */

static void usage(
  const char *pc
) {
  printf("Usage: %s [port]; the port number defaults to %d\n", pc, PORT);
}

#if defined(__sun__) || defined(__osf__)

const char *inet_ntop(
  int         af,
  const void *src,
  char       *dst,
  size_t      cnt
) {
  if (af != AF_INET) {
    errno = EAFNOSUPPORT;
    return 0;
  } else {
    char temp[BUFLEN];
    const unsigned char *puc = (const unsigned char *) src;
    size_t n;

    n = sprintf(temp, "%u.%u.%u.%u", puc[0], puc[1], puc[2], puc[3]);
    if (n >= cnt) {
      errno = ENOSPC;
      return 0;
    } else {
      strcpy(dst, temp);
      return dst;
    }
  }
}

#endif /* __sun__ or __osf__ */
