/* $Id: sockhelp.c,v 1.2 2001/10/18 08:14:29 loreti Exp $ */

#include "sockhelp.h"

#if defined(__linux__) || defined(__sgi__)
int sigemptyset();              /* Black magic?  Prototypes not seen,  */
int sigaction();                /*   also with proper header included. */
#endif

/*-----------------------------------------------------------------------*
 |                       ==> Procedure atoport <==                       |
 |                                                                       |
 | Takes a service name and a service type and returns a port number (or |
 | -1 for error).  If the service name is not found in /etc/services by  |
 | getservbyname(), atoport tries to decode it as a decimal number; the  |
 | value returned is byte ordered for the network.                       |
 *-----------------------------------------------------------------------*/

int atoport(
  const char *service,
  const char *proto
) {
  int             port;
  struct servent *serv;;

  if ((serv = getservbyname((char *) service, (char *) proto)) != 0) {
    port = serv->s_port;
  } else {
    long int  lport;
    char     *errpos;

    lport = strtol(service, &errpos, 0);
    if (*errpos != '\0'   ||   lport < 1   ||   lport > 65535) {
      return -1;
    }
    port = htons(lport);
  }

  return port;
}

/*-----------------------------------------------------------------------*
 |                       ==> Procedure atoaddr <==                       |
 |                                                                       |
 | Converts ASCII string to an "in_addr" structure; NULL is returned if  |
 | the address cannot be found.  The string is first tried as a numeric  |
 | identifier, using inet_addr(); then as an alphanumeric identifier,    |
 | using gethostbyname().                                                |
 *-----------------------------------------------------------------------*/

struct in_addr *atoaddr(
  const char *address
) {
  struct hostent        *host;
  static struct in_addr  saddr;

  if ((saddr.s_addr = inet_addr((char *) address)) != -1) {
    return &saddr;
  } else if ((host = gethostbyname((char *) address)) != 0) {
    return (struct in_addr *) *host->h_addr_list;
  } else {
    return 0;
  }
}

/*-----------------------------------------------------------------------*
 |                    ==> Procedure get_connection <==                   |
 |                                                                       |
 | This function listens on a port, and returns connections (sockets).   |
 | It forks internally, so that your main function doesn't have to worry |
 | about that; this can be confusing if you don't know what is going on. |
 |                                                                       |
 | This function creates a new process for every incoming connection and |
 | so, in the listening process, it will never return.  Only when a      |
 | client connection comes in, and we create a new process to handle it, |
 | the function returns (in the forked child): this means that your code |
 | that calls it MUST *NOT* LOOP.                                        |
 |                                                                       |
 | The calling parameters are as follows:                                |
 | - socket_type (input): SOCK_STREAM or SOCK_DGRAM, for TCP or UDP      |
 |   sockets);                                                           |
 | - port (input): the port to listen on.  Remember that ports<1024 are  |
 |   reserved for the "root" user!  "port" must be in network byte       |
 |   order (see "man htons").                                            |
 | - listener: this is a pointer to a variable that will hold the file   |
 |   descriptor of the socket which is being used to listen.   It is     |
 |   in case you may want to write code to close it, in the event of the |
 |   program termination; but if you are not interested, just pass NULL. |
 |   NOTE: all modern Unix's close file descriptors for you on exit, so  |
 |   that using this variable is usually not required.                   |
 *-----------------------------------------------------------------------*/

int get_connection(
  const int             socket_type,
  const unsigned short  port,
  int                  *listener
) {
  struct sockaddr_in address;
  int                listening_socket;
  int                connected_socket = -1;
  int                reuse_addr = 1;

/**
 | - Checks "socket_type" for valid values;
 | - obtains a socket from the OS;
 | - if "listener" is not NULL returns the listening socket number;
 | - sets SO_REUSEADDR, so that multiple connections may use the same
 |   socket (i.e. the server and all of its forked children).
**/

  if (socket_type != SOCK_STREAM   &&   socket_type != SOCK_DGRAM) {
    fprintf(stderr, "get_connection: invalid socket type = %d\n",
            socket_type);
    exit(EXIT_FAILURE);
  }

  if ((listening_socket = socket(AF_INET, socket_type, 0)) < 0) {
    perror("socket (get_connection)");
    exit(EXIT_FAILURE);
  }

  if (listener != 0) {
    *listener = listening_socket;
  }

  setsockopt(listening_socket, SOL_SOCKET, SO_REUSEADDR,
             (char *) &reuse_addr, sizeof(reuse_addr));

/**
 | Binds a local protocol address to the socket
**/

  memset(&address, 0, sizeof(address));
  address.sin_family      = AF_INET;
  address.sin_port        = port;
  address.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(listening_socket, (SA *) &address, sizeof(address)) < 0) {
    perror("bind (get_connection)");
    close(listening_socket);
    exit(EXIT_FAILURE);
  }

  if (socket_type == SOCK_STREAM) {
    int new_process;

/**
 | - Queue up to five connections before having them automatically
 |   rejected;
 | - call "accept" to wait for the next completed connection for a
 |   client.  If the return value is less than zero, either a real error
 |   occurred, or blocking was interrupted for some reason; in the second
 |   case retries, otherwise aborts the execution.
**/

    listen(listening_socket, 5);

    while (connected_socket < 0) {
      if ((connected_socket = accept(listening_socket, 0, 0)) < 0) {
        if (errno != EINTR) {
          perror("accept (get_connection)");
          close(listening_socket);
          exit(EXIT_FAILURE);
        } else {
          continue;             /* don't fork - do the accept again */
        }
      }

      if ((new_process = fork()) < 0) {

/**
 | Fork error
**/

        perror("fork (get_connection)");
        close(connected_socket);
        connected_socket = -1;

      } else if (new_process == 0) {

/**
 | Child: close our copy of the listening socket
**/

        close(listening_socket);
        if (listener != 0) {
          *listener = -1;
        }
      } else {

/**
 | Parent: this is the main loop.  Close our copy of connected
 | socket, and continue looping.
**/

        close(connected_socket);
        connected_socket = -1;
      }
    }
    return connected_socket;
  } else {
    return listening_socket;
  } /* if (socket-type) */
}

/*-----------------------------------------------------------------------*
 |                   ==> Procedure make_connection <==                   |
 |                                                                       |
 | This is a generic function to make a connection to a given server, or |
 | port.  "service" is a string containing the port name/number; "type"  |
 | is either SOCK_STREAM or SOCK_DGRAM; and "netaddress" is a C string   |
 | containing, in numerical or symbolic form, the host to connect to.    |
 | That function returns the socket, ready for action; or -1 if an error |
 | occurred.                                                             |
 *-----------------------------------------------------------------------*/

int make_connection(
  const char *service,
  const int   type,
  const char *netaddress
) {
  int                 port = -1;
  int                 sock;
  int                 connected;
  struct in_addr     *addr;
  struct sockaddr_in  address;

/**
 | - Checks "type";
 | - gets the port number;
 | - gets the host address;
 | - opens the connection.
**/

  switch (type) {
    case SOCK_STREAM:
      port = atoport(service, "tcp");
      break;

    case SOCK_DGRAM:
      port = atoport(service, "udp");
      break;

    default:
      fputs("make_connection: invalid socket type\n", stderr);
      return -1;
  }

  if (port == -1) {
    fprintf(stderr, "make_connection: invalid service \"%s\"\n", service);
    return -1;
  }

  if ((addr = atoaddr(netaddress)) == 0) {
    fputs("make_connection: invalid network address\n", stderr);
    return -1;
  }

  memset(&address, 0, sizeof(address));
  address.sin_family      = AF_INET;
  address.sin_port        = port;
  address.sin_addr.s_addr = addr->s_addr;

  sock = socket(AF_INET, type, 0);
  fprintf(stdout, "Connecting to %s on port %d ...\n",
          inet_ntoa(*addr), ntohs(port));

  if (type == SOCK_STREAM) {
    if ((connected = connect(sock, (SA *) &address, sizeof(address))) < 0) {
      perror("connect (make_connection)");
      return -1;
    }
    return sock;
  }

/**
 | If here, must be for UDP: so, bind to address.
**/

  if (bind(sock, (SA *) &address, sizeof(address)) < 0) {
    perror("bind (make_connection");
    return -1;
  }
  return sock;
}

/*-----------------------------------------------------------------------*
 |                      ==> Procedure sock_read <==                      |
 |                                                                       |
 | This is just like the read() system call, except that it will make    |
 | sure that all your data goes through the socket.                      |
 *-----------------------------------------------------------------------*/

int sock_read(
  const int     sockfd,
  char         *buf,
  const size_t  count
) {
  size_t bytes_read = 0;
  int    this_read;

  while (bytes_read < count) {
    do {
      this_read = read(sockfd, buf, count-bytes_read);
    } while (this_read < 0   &&   errno == EINTR);

    if (this_read < 0) {
      return this_read;
    } else if (this_read == 0) {
      return bytes_read;
    }

    bytes_read += this_read;
    buf        += this_read;
  }

  return count;
}

/*-----------------------------------------------------------------------*
 |                      ==> Procedure sock_write <==                     |
 |                                                                       |
 | This is just like the write() system call, except that it will make   |
 | sure that all data is transmitted.                                    |
 *-----------------------------------------------------------------------*/

int sock_write(
  const int     sockfd,
  const char   *buf,
  const size_t  count
) {
  size_t bytes_sent = 0;
  int    this_write;

  while (bytes_sent < count) {
    do {
      this_write = write(sockfd, buf, count-bytes_sent);
    } while (this_write < 0   &&   errno == EINTR);

    if (this_write <= 0) {
      return this_write;
    }

    bytes_sent += this_write;
    buf        += this_write;
  }

  return count;
}

/*-----------------------------------------------------------------------*
 |                      ==> Procedure sock_gets <==                      |
 |                                                                       |
 | This function reads from a socket, until it receives a line-feed (\n) |
 | character.  It fills the buffer "str" up to the maximum size "count"; |
 | or will return -1 if the socket is closed during the read operation.  |
 |                                                                       |
 | NOTE that if a single line exceeds the length of "count", the extra   |
 | data will be READ AND DISCARDED.  You have been warned.               |
 *-----------------------------------------------------------------------*/

int sock_gets(
  const int     sockfd,
  char         *str,
  const size_t  count
) {
  int   bytes_read;
  int   total_count      = 0;
  char *current_position = str;
  char  last_read        = '\0';

  while (last_read != '\n') {
    if ((bytes_read = read(sockfd, &last_read, 1)) <= 0) {
      return -1;
    }

    if (total_count < count   &&   last_read != '\n') {
      *current_position++ = last_read;
      total_count++;
    }
  }

  if (count > 0) {
    *current_position = '\0';
  }

  return total_count;
}

/*-----------------------------------------------------------------------*
 |                      ==> Procedure sock_puts <==                      |
 |                                                                       |
 | This function writes a character string out to a socket.  It will     |
 | return -1 if the connection is closed while it is trying to write.    |
 *-----------------------------------------------------------------------*/

int sock_puts(
  const int   sockfd,
  const char *str
) {
  return sock_write(sockfd, str, strlen(str));
}

/*-----------------------------------------------------------------------*
 |                      ==> Procedure my_signal <==                      |
 |                                                                       |
 | Our own interface with "sigaction".  See: W. Richard Stevens, "Unix   |
 | Network Programming", Vol. 1, pages 120 and following for details;    |
 | this is essentially the (obsolete) procedure "signal", that predated  |
 | POSIX.1 .                                                             |
 |                                                                       |
 | Under Solaris 2.6 "sigaction" does not work; but there is a working   |
 | implementation of "signal".                                           |
 *-----------------------------------------------------------------------*/

#if defined(__sun__)

sigFunction *my_signal(
  const int sigNr,
  sigFunction *fun
) {
  return signal(sigNr, fun);
}

#else

sigFunction *my_signal(
  const int sigNr,
  sigFunction *fun
) {
  struct sigaction action, oldAction;

/**
 | - Declare "fun" as handler for "sigNr": no additional signals to
 |   be blocked when our signal handler is called;
 | - patch for SA_INTERRUPT (bad behavior of older SunOS 4.x);
 | - patch for SA_RESTART (SVR4, 4.4BSD).
**/

  sigemptyset(&action.sa_mask);
  action.sa_handler = fun;
  action.sa_flags   = 0;

  if (sigNr == SIGALRM) {
#if defined(SA_INTERRUPT)
    action.sa_flags |= SA_INTERRUPT;
#endif
  } else {
#if defined(SA_RESTART)
    action.sa_flags |= SA_RESTART;
#endif
  }

  if (sigaction(sigNr, &action, &oldAction) < 0) {
    return SIG_ERR;
  } else {
    return oldAction.sa_handler;
  }
}

#endif /* __sun__ */
